#############################################################################
##
#W    complexes.gi        share package 'homology'          Heckenbach/Welker
##
## Some example complexes and basic operations on simplicial complexes.

#==================================================
# Example simplicial complexes
#==================================================

InstallGlobalFunction (SCSphere, function (n)
  return List ([1 .. n + 2], i -> Difference ([1 .. n + 2], [i]));
end);

InstallValue (SCMobiusStrip,
  [["a", "d", "e"], ["a", "b", "e"], ["b", "c", "e"],
   ["c", "e", "f"], ["c", "d", "f"], ["a", "d", "f"]]);

InstallValue (SCTorus,
  [["a", "b", "e"], ["b", "e", "d"], ["b", "d", "f"],
   ["b", "c", "f"], ["a", "d", "f"], ["c", "e", "f"],
   ["a", "c", "e"], ["a", "f", "g"], ["f", "g", "e"],
   ["e", "d", "g"], ["a", "c", "d"], ["a", "b", "g"],
   ["b", "g", "c"], ["c", "g", "d"]]);

InstallGlobalFunction (SCChessboard, function (m, n)
  local f;
  f := function (s, n)
    if n < 1 then
      return [[]];
    else
      return Concatenation (List (s, i -> List (f (Difference (s, [i]), n - 1),
        a -> Concatenation (a, [[i, n]]))));
    fi;
  end;
  if m < n then
    return f ([1 .. n], m);
  else
    return f ([1 .. m], n);
  fi;
end);

InstallGlobalFunction (SCMatching, function(n)
  local Matching_list, all_matchings;
  Matching_list := function(debut, remain)   
    local i, j, s;
    if Size(remain) >= 2 then
       s := 0;
       if Size(debut) > 0 then
           s := debut[Size(debut)][1];
       fi;
       for i in [1 .. Size(remain)] do
         if (remain[i] > s) then
           for j in [(i+1) .. Size(remain)] do
             Matching_list( Concatenation(debut,[[remain[i],remain[j]]]),
                            Filtered(remain,x->(x<>remain[i] and x<>remain[j])));
           od;
         fi;
       od; 
    else
           all_matchings := Concatenation(all_matchings, [debut]);
    fi;
  end;
  all_matchings:=[];
  Matching_list([],[1..n]);
  return all_matchings;
end);



InstallGlobalFunction (SCGeneralizedKleinBottle, function (n)
  local Strip, Filler;
  Strip := function (a1, a2, b, c, d, e, f)
    return [[a1, b, d], [e, b, d], [e, b, c], [e, f, c], [a2, f, c], [a2, f, d]];
  end;
  Filler := function (i, j, k)
    return List ([i .. j - 1], l -> [l, l + 1, k]);
  end;
  return Concatenation (
    Strip ("a", "a", "b", "c", "d", "e", "f"),
    Strip ("d", "d", "e", "f", 0, n - 1, 2 * (n - 1)),
    Concatenation (List ([0, 3 .. 3 * n - 9], i ->
      Strip (i, i + 3, i + 1, i + 2, "a", "c", "b"))),
    Strip (3 * n - 6, 0, 3 * n - 5, 3 * n - 4, "a", "c", "b"),
    Filler (0, n - 2, n - 1),
    Filler (n - 1, 2 * n - 3, 2 * n - 2),
    Filler (2 * n - 2, 3 * n - 4, 0));
end);

InstallValue (SCKleinBottle, SCGeneralizedKleinBottle (2));

InstallGlobalFunction (SCNot2ConnectedGraphs, function (n)
  return Concatenation (List ([1 .. n],
    i -> List (Difference (Combinations (Difference ([1 .. n], [i, i mod n + 1])), [[]]),
      s -> List (Concatenation (
        List (Difference ([1 .. n], [i]), j -> [i, j]),
        Combinations (s, 2),
        Combinations (Difference (Difference ([1 .. n], [i]), s), 2)), AsSortedList))));
end);

#==================================================
# Operations on simplicial complexes
#==================================================

InstallGlobalFunction (SCWedge, function (Complexes)
  local Points;
  if not (IsList (Complexes) and ForAll (Complexes, IsSimplicialComplex)) then
    Info (InfoWarning, 1, "SCWedge: the argument should be a list of simplicial complexes");
    return fail;
  fi;
  Points := List (Complexes, c -> Union (c));
  return Concatenation (List (Filtered ([1 .. Length (Complexes)], i -> Points [i] <> []),
    i -> List (Complexes [i], s -> List (s,
      function (n)
        if n = Points [i] [1] then
          return "a";
        else
          return [i, n];
        fi;
      end))));
end);

InstallGlobalFunction (SCJoin, function (Complex1, Complex2)
  if not (IsSimplicialComplex (Complex1) and IsSimplicialComplex (Complex2)) then
    Info (InfoWarning, 1, "SCJoin: the arguments should be simplicial complexes");
    return fail;
  fi;
  return Concatenation (List (Complex1, s1 -> List (Complex2,
    s2 -> Concatenation (List (s1, i -> [1, i]), List (s2, i -> [2, i])))));
end);

InstallGlobalFunction (SCCone, function (Complex)
  if not IsSimplicialComplex (Complex) then
    Info (InfoWarning, 1, "SCCone: the argument should be a simplicial complex");
    return fail;
  fi;
  return SCJoin (Complex, [[1]]);
end);

InstallGlobalFunction (SCSuspension, function (Complex)
  if not IsSimplicialComplex (Complex) then
    Info (InfoWarning, 1, "SCSuspension: the argument should be a simplicial complex");
    return fail;
  fi;
  return SCJoin (Complex, [[1], [2]]);
end);

InstallGlobalFunction (SCStar, function (Complex,Face)
  if not IsSimplicialComplex (Complex) then
    Info (InfoWarning, 1, "SCStar: the argument should be a simplicial complex");
    return fail;
  fi;
  return Filtered(Complex, x -> IsSubset(x,Face));
end);

InstallGlobalFunction (SCLink, function (Complex,Face)
  if not IsSimplicialComplex (Complex) then
    Info (InfoWarning, 1, "SCLink: the argument should be a simplicial complex");
    return fail;
  fi;
  return List(SCStar(Complex,Face), x -> Difference(x,Face));
end);

InstallGlobalFunction (SCCompactify, function (Complex)
  if not IsSimplicialComplex (Complex) then
    Info (InfoWarning, 1, "SCCompactify: the argument should be a simplicial complex");
    return fail;
  fi;
  
  Complex := DuplicateFreeList(Complex);
  return Filtered(Complex, Face1 -> Number(Complex, Face2 -> IsSubset(Face2,Face1)) = 1);
end);

InstallGlobalFunction (SCRestriction, function (Complex,Face)
  if not IsSimplicialComplex (Complex) then
    Info (InfoWarning, 1, "SCRestriction: the argument should be a simplicial complex");
    return fail;
  fi;
  return List(Complex, x -> Intersection(x,Face));
end);

InstallGlobalFunction (SCDimension, function(Complex)
  if not IsSimplicialComplex (Complex) then
    Info (InfoWarning, 1, "SCRestriction: the argument should be a simplicial complex");
    return fail;
  fi;

  return MaximumList(List(Complex,set->Number(set)))-1;
end);

InstallGlobalFunction (SCAlexanderDual, function(Complex)
  local ground_set;

  if not IsSimplicialComplex (Complex) then
    Info (InfoWarning, 1, "SCRestriction: the argument should be a simplicial complex");
    return fail;
  fi;
 
  ground_set := Union(Complex); 
  return List(SCMinimalNonFaces(Complex), non_face -> Difference(ground_set,non_face));

end);

InstallGlobalFunction (SCMinimalNonFaces, function(Complex)
  local ground_set,i,i_set, subfaces, non_faces,face,t_face,x_face,x,one_subsets,face_complement,sections,full_x_skeleton; 

  if not IsSimplicialComplex (Complex) then
    Info (InfoWarning, 1, "SCRestriction: the argument should be a simplicial complex");
    return fail;
  fi;

  non_faces := [];
  ground_set := Union(Complex);

  i := MinimumList([SCDimension(Complex)+2,Number(ground_set)]); 


  for face in Complex do
    i := Number(face);
    face_complement := Difference(ground_set,face);
    one_subsets := Combinations(face,1);
    while (i >= 0) do
      subfaces := Combinations(face,i);
      for x in  face_complement do
        full_x_skeleton := true;
        for i_set in subfaces do 
          x_face := Union(i_set,[x]); 
          sections := List(Complex,t_face
                                      -> Difference(x_face,t_face));
          if not([] in sections) then   
            full_x_skeleton := false; 
            if Number(Intersection(sections ,Union (one_subsets,[[x]]))) 
                    = i+1 then
              Append(non_faces,[x_face]);
            fi;
          fi;
        od;
        if full_x_skeleton = true then
          face_complement := Difference(face_complement,[x]);
        fi;
      od;
      i:=i-1;
    od;
  od;

  return(SCCompactify(non_faces));
end);
  
InstallGlobalFunction (SCDeletedJoin, function(Complex1,Complex2)
  if not (IsSimplicialComplex (Complex1) and IsSimplicialComplex (Complex2)) then
    Info (InfoWarning, 1, "SCRestriction: the argument should be a simplicial complex");
    return fail;
  fi;

  Complex1 := Union(List(Complex1, face -> Combinations(face)));
  Complex2 := Union(List(Complex2, face -> Combinations(face)));

  return SCCompactify(
            List(
              Filtered(
                 Union(
                   List(Complex1, face1 -> 
                                     List(Complex2, face2 -> [face1,face2]))), 
                 pair -> IsEmpty(Intersection(pair[1],pair[2]))), 
             pair -> Union(List(pair[1],e -> [e,1]),List(pair[2],e -> [e,2]))));
end);
